#include "utils/encodeDecode.glsl"
#include "utils/mipmap.glsl"
#include "utils/project.glsl"
#include "utils/rand.glsl"
#include "utils/readDepth.glsl"

out vec4 oFragColor;

#ifdef SSAO_NORMAL
uniform sampler2D uTextureGbuffer2;
#endif

#ifdef DEPTH_MIPMAP
uniform sampler2D uTextureDepthMipmap;
uniform vec2 uTextureDepthMipmapSize;
#else
uniform sampler2D uTextureDepth;
#endif

uniform float uQuality; // mipmap usage
uniform float uFrameJitter;

uniform float uSceneScale;

uniform float uSsaoWorldRadius;
uniform float uSsaoFactor;
uniform float uSsaoBias;
uniform float uSsaoScreenRadius;
uniform vec4 uProjectionScaleBias;
uniform vec4 uProjectionDepth;

// Constant used to scale the depth view value that is given to the blur pass
#define SSAO_JITTER 1.0
#define MIN_RADIUS 1.0
#define NB_SPIRAL_TURNS 3.0

#define MAX_DEPTH_LOD 5.0

// lower value => faster
#define DEPTH_OFFSET_LEVEL 3.0

#define MAX_DEPTH_LEVEL 5.0
#define MIN_DEPTH_LEVEL 0.0

float fetchDepthPackLevel(const in vec2 uv, const in float level) {
#ifdef DEPTH_MIPMAP
    return texture(uTextureDepthMipmap, computeLodUV(uv, level, uTextureDepthMipmapSize.x)).r;
#else
    return readDepthLinear(texture(uTextureDepth, uv).r);
#endif
}

vec3 getOffsetViewPosition(const in vec2 unitOffset, const in float screenRadius) {
    vec2 uv = g_uv + floor(unitOffset * screenRadius) / uOutputSize;

    if (uv.x > 1.0 || uv.x < 0.0 || uv.y > 1.0 || uv.y < 0.0) {
        return vec3(0.0);
    }

    // if (uv.x > 1.0) uv.x = 2.0 - uv.x;
    // if (uv.y > 1.0) uv.y = 2.0 - uv.y;
    // if (uv.x < 0.0) uv.x = -uv.x;
    // if (uv.y < 0.0) uv.y = -uv.y;

    float mipLevel = floor(log2(screenRadius)) - DEPTH_OFFSET_LEVEL;
    mipLevel = clamp(mipLevel, 0.0, MAX_DEPTH_LEVEL);

    float depth = fetchDepthPackLevel(uv, mipLevel);
    return vec3(uv * uProjectionScaleBias.xy + uProjectionScaleBias.zw, 1.0) * -depthLinearToZEye(depth);
}

float tapLoop11(const in vec3 eyeVector, const in vec3 normal, const in float screenRadius, const in float worldRadius2,
                const in float randomAngle, const in float invNumSamples, const in int offset) {
    float contrib = 0.0;

    float epsilon = uSceneScale * 0.01;
    epsilon = epsilon * epsilon;

    for (int i = 0; i < 11; ++i) {
        float sradius = (float(offset + i) + 0.5) * invNumSamples;
        float angle = sradius * (NB_SPIRAL_TURNS * 6.28) + randomAngle;

        vec3 occludePoint = getOffsetViewPosition(vec2(cos(angle), sin(angle)), sradius * screenRadius);
        if (occludePoint.z == 0.0) {
            continue;
        }

        occludePoint -= eyeVector;

        float vv = dot(occludePoint, occludePoint);
        float vn = dot(occludePoint, normal) - uSsaoBias;

        float f = max(1.0 - vv / worldRadius2, 0.0);

        // https://github.com/PeterTh/gedosato/blob/master/pack/assets/dx9/SAO.fx#L320
        contrib += f * max(vn * inversesqrt(epsilon + vv), 0.0);

        // https://gist.github.com/transitive-bullshit/6770311#file-billboard_sao-frag-L72
        // contrib += f * f * f * max(vn / (epsilon + vv), 0.0);
    }

    return contrib;
}

void main(void) {
    initUV();

    float depth = fetchDepthPackLevel(g_uv, 0.0);
    vec3 eyeVector = vec3(g_uv * uProjectionScaleBias.xy + uProjectionScaleBias.zw, 1.0) * -depthLinearToZEye(depth);

#ifndef SSAO_NORMAL
    vec3 normal = cross(dFdy(eyeVector), dFdx(eyeVector));
    // vec3 depthLeft = getOffsetViewPosition(vec2(-1.0, 0.0), 1.0);
    // vec3 depthRight = getOffsetViewPosition(vec2(1.0, 0.0), 1.0);
    // vec3 depthBottom = getOffsetViewPosition(vec2(0.0, -1.0), 1.0);
    // vec3 depthTop = getOffsetViewPosition(vec2(1.0, 1.0), 1.0);
    // vec3 normal = cross(depthLeft - depthRight, depthTop - depthBottom);
#endif

    // early returns should be done after derivatives usage to avoid bad edges
    // early return background
    if (depth > 0.99) {
        oFragColor = vec4(1.0);
        return;
    }

    float screenRadius = uSsaoScreenRadius / max(eyeVector.z, 0.0001);

    // early return radius too small (note: needs to be in separate conditional because of nvidia/osx bug)
    if (screenRadius < MIN_RADIUS) {
        oFragColor = vec4(1.0);
        return;
    }

#ifdef SSAO_NORMAL
    vec3 normal = -decodeNormal(texture(uTextureGbuffer2, g_uv).rg);
#else
    normal = normalize(normal);
#endif

    // ------------------------------
    float nbSamples = 11.0;
    if (uQuality > 0.33) nbSamples += 11.0;
    if (uQuality > 0.66) nbSamples += 11.0;

    float invNumSamples = 1.0 / nbSamples;
    float randomAngle = SSAO_JITTER * 2.0 * M_PI * interleaveGradientNoise(gl_FragCoord.xy, uFrameJitter);
    float worldRadius2 = uSsaoWorldRadius * uSsaoWorldRadius;

    // ------------------------------
    float contrib = 0.0;
    contrib += tapLoop11(eyeVector, normal, screenRadius, worldRadius2, randomAngle, invNumSamples, 0);
    if (uQuality > 0.33)
        contrib += tapLoop11(eyeVector, normal, screenRadius, worldRadius2, randomAngle, invNumSamples, 11);
    if (uQuality > 0.66)
        contrib += tapLoop11(eyeVector, normal, screenRadius, worldRadius2, randomAngle, invNumSamples, 22);

    // ------------------------------
    float nFalloff = mix(1.0, max(0.0, 1.5 * normal.z), 0.35);
    float aoValue = max(0.0, 1.0 - sqrt(contrib * nFalloff * invNumSamples));
    aoValue = pow(aoValue, uSsaoFactor * 30.0);

    aoValue = mix(1.0, aoValue, saturate(screenRadius - MIN_RADIUS));
    oFragColor = vec4(aoValue, depth, 0.0, 0.0);
}
