#include "utils/encodeDecode.glsl"
#include "utils/mipmap.glsl"
#include "utils/project.glsl"
#include "utils/rand.glsl"
#include "utils/readDepth.glsl"

out vec4 oFragColor;

uniform sampler2D uTexture0;
uniform sampler2D uTextureGbuffer0;
uniform sampler2D uTextureGbuffer2;

uniform vec2 uTexture0Size;

uniform float uFrameJitter;
uniform float uQuality;
uniform vec4 uProjectionScaleBias;
uniform vec2 uNearFar;
uniform mat4 uReprojectUV;

uniform float uSsrDistanceFading;

#ifdef DEPTH_MIPMAP
uniform sampler2D uTextureDepthMipmap;
uniform vec2 uTextureDepthMipmapSize;
#else
uniform sampler2D uTextureDepth;
#endif

// #define DEBUG_RAY
#ifdef DEBUG_RAY
vec2 g_uv0 = vec2(0.45, 0.25);
vec3 g_colorDebug = vec3(0.0);
#endif

#define SSR_STEPS 16

#define LOD_COLOR 5.0
#define LOD_DEPTH 3.0

vec3 fetchColorLod(const in vec2 uv, const in float level) {
    return texture(uTexture0, computeLodUV(uv, level * LOD_COLOR, uTexture0Size.x)).rgb;
}

float fetchDepthLod(const in vec2 uv, const in vec3 lodParams) {
#ifdef DEPTH_MIPMAP
    float depth = texture(uTextureDepthMipmap, computeLodUV(uv, lodParams)).r;
#else
    float depth = readDepthLinear(texture(uTextureDepth, uv).r);
#endif

    if (depth >= 0.999) return 1000.0; // very far so it won't reflect background
    return depth;
}

vec3 ssrViewToUV(const in vec3 viewPos) {
    vec2 uv = -viewPos.xy / (uProjectionScaleBias.zw * viewPos.z);
    return vec3(0.5 + 0.5 * uv, zEyeToDepthNoLinear(viewPos.z));
}

vec3 tapLoop8(const in vec3 diffIn, const in vec3 rayOriginUV, const in vec3 rayDirUV, const in float depthTolerance,
              const in float invNumSteps, const in vec3 depthLodParams) {
    vec3 diffSample = diffIn;

#ifdef DEBUG_RAY
    if (distance(g_uv0.xy, rayOriginUV.xy) < 0.01) {
        g_colorDebug = vec3(1.0, 0.0, 0.0);
    }
#endif

    for (int i = 0; i < SSR_STEPS; ++i) {
        vec3 sampleUV = rayOriginUV + rayDirUV * diffSample.y;

#ifdef DEBUG_RAY
        if (distance(g_uv0, sampleUV.xy) < 0.01) {
            g_colorDebug = vec3(1.0, 1.0, 1.0);
            if (i == 0) g_colorDebug = vec3(0.0, 1.0, 0.0);
            if (i == 1) g_colorDebug = vec3(0.0, 0.0, 1.0);
            if (i == 2) g_colorDebug = vec3(1.0, 1.0, 0.0);
            if (i == 3) g_colorDebug = vec3(0.0, 1.0, 1.0);
            if (i == SSR_STEPS - 3) g_colorDebug = vec3(1.0, 0.0, 1.0);
            if (i == SSR_STEPS - 2) g_colorDebug = vec3(0.0, 1.0, 0.0);
            if (i == SSR_STEPS - 1) g_colorDebug = vec3(0.0, 0.0, 1.0);
        }
#endif
        float depthDiff = -toDepthLinear(sampleUV.z) + fetchDepthLod(sampleUV.xy, depthLodParams);

        float hitTime = 1.0;

        // ]-2 * depthTolerance, 0[
        if (abs(depthDiff + depthTolerance) < depthTolerance) {
            hitTime = diffSample.y + invNumSteps * (saturate(diffSample.x / (diffSample.x - depthDiff)) - 1.0);
        }

        diffSample.x = depthDiff;
        diffSample.y += invNumSteps;
        diffSample.z = min(diffSample.z, hitTime);
    }

    return diffSample;
}

vec3 clampUV(const in vec3 rayOriginUV, const in vec3 rayDirUV) {
    // clip the line to the viewport edges
    // https://github.com/godotengine/godot/blob/master/drivers/gles3/shaders/screen_space_reflection.glsl#L124
    float scaleMaxX = min(1.0, 0.99 * (1.0 - rayOriginUV.x) / max(1e-5, rayDirUV.x));
    float scaleMaxY = min(1.0, 0.99 * (1.0 - rayOriginUV.y) / max(1e-5, rayDirUV.y));
    float scaleMinX = min(1.0, 0.99 * rayOriginUV.x / max(1e-5, -rayDirUV.x));
    float scaleMinY = min(1.0, 0.99 * rayOriginUV.y / max(1e-5, -rayDirUV.y));
    return rayDirUV * (min(scaleMaxX, scaleMaxY) * min(scaleMinX, scaleMinY));
}

vec4 rayTraceUnrealSimple(const in vec3 rayOriginUV, const in vec3 rayEndUV, const in float rayLen,
                          const in float roughness, const in float jitter) {

#ifdef DEPTH_MIPMAP
    vec3 lodParams = computeLodParams(roughness * LOD_DEPTH, uTextureDepthMipmapSize.x);
#else
    vec3 lodParams = vec3(0.0);
#endif

    float invNumSteps = 1.0 / float(SSR_STEPS);
    float depthTolerance = 0.5 * invNumSteps * rayLen / (uNearFar.y - uNearFar.x);

    vec3 rayDirUV = clampUV(rayOriginUV, rayEndUV);

    float rnd = interleaveGradientNoise(gl_FragCoord.xy, uFrameJitter * 2.0 + jitter);

    vec3 diffSample; // [depthDiff, ray_advance 0-1, tRayHit]
    diffSample.x = 0.0;
    diffSample.y = invNumSteps * (0.5 + rnd);
    diffSample.z = 1.0;

    // ---------------------

    diffSample = tapLoop8(diffSample, rayOriginUV, rayDirUV, depthTolerance, invNumSteps, lodParams);

    return vec4(rayOriginUV + rayDirUV * diffSample.z, 1.0 - diffSample.z);
}

vec4 fetchColorContribution(const in vec4 resRay, const in float roughness) {
    vec4 tmp = uReprojectUV * vec4(resRay.xyz * 2.0 - 1.0, 1.0);
    vec2 uv = tmp.xy / tmp.w;

    // fade edge of screen
    float maskEdge = clamp(6.0 - 6.0 * max(abs(uv.x), abs(uv.y)), 0.0, 1.0);
    uv = 0.5 + 0.5 * uv;

    maskEdge *= mix(1.0, resRay.w, uSsrDistanceFading);
    // return vec4(fetchColorLod(uv, roughness * clamp((1.0 - resRay.w), 0.0, 1.0)), maskEdge);
    return vec4(fetchColorLod(uv, roughness), maskEdge);
}

// https://github.com/EpicGames/UnrealEngine/blob/release/Engine/Shaders/Private/ScreenSpaceReflections.usf#L143
// https://github.com/EpicGames/UnrealEngine/blob/release/Engine/Shaders/Private/MonteCarlo.ush#L169
vec3 unrealImportanceSampling(const in vec3 tX, const in vec3 tY, const in vec3 tZ, const in vec3 eyeVector,
                              const in float rough4, const in float jitter) {
    vec2 E;

    E.x = interleaveGradientNoise(gl_FragCoord.yx, uFrameJitter * 3.0 + jitter);

    // uniform vec3 uHalton;
    // uniform sampler2D uTextureBlueNoise;
    // E.x = texture(uTextureBlueNoise, gl_FragCoord.xy / 64.0 + uHalton.xy + jitter * 0.5).x;

    E.y = fract(E.x * 52.9829189);

    // Reduces BRDF tail is noisy, we clamp it
    // https://www.slideshare.net/DICEStudio/stochastic-screenspace-reflections#p=67
    E.y = mix(E.y, 1.0, 0.7);

    float phi = (M_PI * 2.0) * E.x;
    // float cosTheta = sqrt((1.0 - E.y) / (1.0 + (rough4 - 1.0) * E.y)); // ggx (long tail, more noisy)
    float cosTheta = pow(max(E.y, 0.000001), rough4 / (2.0 - rough4)); // blinn
    float sinTheta = sqrt(1.0 - cosTheta * cosTheta);

    vec3 h = vec3(sinTheta * cos(phi), sinTheta * sin(phi), cosTheta);
    h = h.x * tX + h.y * tY + h.z * tZ;
    return reflect(-eyeVector, h);
}

vec4 castRay(const in vec3 rayOriginUV, const in vec2 rayRange, const in vec3 tX, const in vec3 tY,
             const in vec3 normal, const in vec3 eyeVector, const in vec3 rayStart, const in float roughness,
             const in float rough4, const in float jitter) {
    // direction of reflection
    vec3 rayDirView = unrealImportanceSampling(tX, tY, normal, eyeVector, rough4, jitter);
    float rayLen = mix(rayRange.x, rayRange.y, rayDirView.z * 0.5 + 0.5);

    if (dot(rayDirView, normal) > 0.001) {
        vec3 rayEndUV = ssrViewToUV(rayStart + rayDirView * rayLen) - rayOriginUV;
        vec4 resRay = rayTraceUnrealSimple(rayOriginUV, rayEndUV, rayLen, roughness, jitter);
        if (resRay.w > 0.0) {
            return fetchColorContribution(resRay, roughness);
        }
    }

    return vec4(0.0);
}

vec4 ssr(const in float roughness, const in vec3 normal, const in vec3 eyeVector, const in vec3 rayStart) {
    // precompute importance sampling roughness^4
    float rough4 = roughness * roughness;
    rough4 = rough4 * rough4;

    // precompute importance sampling basis tangent
    vec3 upVector = abs(normal.z) < 0.999 ? vec3(0.0, 0.0, 1.0) : vec3(1.0, 0.0, 0.0);
    vec3 tX = normalize(cross(upVector, normal));
    vec3 tY = cross(normal, tX);

    // fade ray that points toward camera
    float maskSsr = clamp(-4.0 * dot(eyeVector, normal) + 3.8, 0.0, 1.0);
    maskSsr *= clamp(4.7 - roughness * 5.0, 0.0, 1.0); // fade out ssr for high roughness values

    if (maskSsr <= 0.0) {
        return vec4(0.0);
    }

    // compute ray uv origin
    vec3 rayOriginUV = ssrViewToUV(rayStart);

    // if reflection is toward us, we cast [near, vertex], otherwise [vertex, far]
    vec2 rayRange = vec2(uNearFar.y + rayStart.z, -rayStart.z - uNearFar.x);

    // w: fail at 0, hit at 1
    vec4 result = castRay(rayOriginUV, rayRange, tX, tY, normal, eyeVector, rayStart, roughness, rough4, 0.0);

    if (uQuality > 0.66) {
        result += castRay(rayOriginUV, rayRange, tX, tY, normal, eyeVector, rayStart, roughness, rough4, 1.0);
        result /= 2.0;
    }

    result.w *= maskSsr;
    return result;
}

void main(void) {
    initUV();

    vec4 buf0 = texture(uTextureGbuffer0, g_uv);
    if (buf0.a == 0.0) {
        discard;
    }

#ifdef DEBUG_RAY
    g_uv = g_uv0;
    g_uv0 = gl_FragCoord.xy / uOutputSize;
#endif

    float roughness = buf0.a;
    vec3 normal = decodeNormal(texture(uTextureGbuffer2, g_uv).rg);
    vec3 eyeVector = vec3(g_uv * uProjectionScaleBias.xy + uProjectionScaleBias.zw, 1.0);

    // ---------------------

#ifdef DEPTH_MIPMAP
    float depth = texture(uTextureDepthMipmap, computeLodUV(g_uv, 0.0, uTextureDepthMipmapSize.x)).r;
#else
    float depth = readDepthLinear(texture(uTextureDepth, g_uv).x);
#endif

    vec3 rayStart = eyeVector * depthLinearToZEye(depth);
    oFragColor = ssr(roughness, normal, eyeVector, rayStart);

#ifdef DEBUG_RAY
    oFragColor.rgb = g_colorDebug;
    oFragColor.w = 1.0;
#endif
}
