uint ReverseBits32(uint bits) {
    bits = (bits << 16U) | ( bits >> 16U);
    bits = ((bits & 0x00ff00ffU) << 8U) | ((bits & 0xff00ff00U) >> 8U);
    bits = ((bits & 0x0f0f0f0fU) << 4U) | ((bits & 0xf0f0f0f0U) >> 4U);
    bits = ((bits & 0x33333333U) << 2U) | ((bits & 0xccccccccU) >> 2U);
    bits = ((bits & 0x55555555U) << 1U) | ((bits & 0xaaaaaaaaU) >> 1U);
    return bits;
}

vec2 Hammersley( uint Index, uint NumSamples, uvec2 Random) {
    float E1 = fract(float(Index) / float(NumSamples) + float(Random.x & 0xffffU) / float(1<<16));
    // float E1 = fract((float)Index / NumSamples + float(Random.x & 0xffffU) / (1<<16));
    float E2 = float(ReverseBits32(Index) ^ Random.y) * 2.3283064365386963e-10;
    return vec2(E1, E2);
}

uvec2 Random = uvec2(pseudoRandom(gl_FragCoord.xy + frameMod * vec2(97.0, 71.0)) * vec2(uvec2(0x3127352U, 0x11229256U)));
E = Hammersley(uint(i), uint(SSR_NUM_RAY), Random);

vec4 rayTraceUnrealVectorized(
    const in vec3 viewVertex, const in vec3 rayDirView, const in mat4 projection,
    const in sampler2D depthTex, const in vec2 ratio, const in vec2 nearFar, const in float boxRadius, const in vec4 halton,
    const in float roughness, const in vec2 size, const in float frameMod) {

    float invNumSteps = 1.0 / float(SSR_STEPS);

    vec3 rayOriginUV, rayDirUV;
    float depthTolerance = computeScreenRayOriginDir(viewVertex.xyz, rayDirView, nearFar, projection, rayDirUV, rayOriginUV);
    vec4 result = vec4(rayOriginUV, 1.0);

    rayDirUV *= invNumSteps;
    vec3 rayUVz = rayOriginUV + rayDirUV * getStepOffset(frameMod);

    float lastDiff = 0.0;
    float level = HZB_LEVEL_OFFSET;

    for (int i = 0; i < SSR_STEPS; i += 4)
    {
        // Vectorized to group fetches
        vec4 sampleUV0 = rayUVz.xyxy + rayDirUV.xyxy * vec4(1.0, 1.0, 2.0, 2.0);
        vec4 sampleUV1 = rayUVz.xyxy + rayDirUV.xyxy * vec4(3.0, 3.0, 4.0, 4.0);
        vec4 sampleZ   = rayUVz.zzzz + rayDirUV.zzzz * vec4(1.0, 2.0, 3.0, 4.0);
        
        // Use lower res for farther samples
        vec4 sampleDepth = fetchDepth(depthTex, sampleUV0, sampleUV1, nearFar, level, size, ratio);

        vec4 depthDiff = -1.0 / sampleZ - sampleDepth;
        bvec4 hit;
        hit.x = abs(depthDiff.x + depthTolerance) < depthTolerance;
        hit.y = abs(depthDiff.y + depthTolerance) < depthTolerance;
        hit.z = abs(depthDiff.z + depthTolerance) < depthTolerance;
        hit.w = abs(depthDiff.w + depthTolerance) < depthTolerance;

        if (hit.x || hit.y || hit.z || hit.w) {
            float depthDiff0 = depthDiff[2];
            float depthDiff1 = depthDiff[3];
            float minTime = 3.0;

            if (hit[2]) {
                depthDiff0 = depthDiff[1];
                depthDiff1 = depthDiff[2];
                minTime = 2.0;
            }

            if (hit[1]) {
                depthDiff0 = depthDiff[0];
                depthDiff1 = depthDiff[1];
                minTime = 1.0;
            }

            if (hit[0]) {
                depthDiff0 = lastDiff;
                depthDiff1 = depthDiff[0];
                minTime = 0.0;
            }

            // Find more accurate hit using line segment intersection
            float timeLerp = clamp(depthDiff0 / (depthDiff0 - depthDiff1), 0.0, 1.0);
            float intersectTime = minTime + timeLerp;

            result.xyz = rayUVz + rayDirUV * intersectTime;
            result.w = 0.5;
            break;
        }

        level += roughness * (16.0 * invNumSteps);
        lastDiff = depthDiff.w;
        rayUVz += 4.0 * rayDirUV;
    }


    result.w = 1.0 - result.w;
    return result;
}

vec4 rayTraceMarmoset(mat4 projection, sampler2D depthTex, vec2 nearFar, vec3 pos, vec3 dd, vec2 size, vec2 ratio) {
    vec3 dir = dd;

    // ray march settings
    float maxRayDist = abs(pos.z) / sqrt(0.01 + clamp(dot(dir.xy, dir.xy), 0.0, 1.0));
    float incr = maxRayDist / float(SSR_STEPS);
    
    bool under = false, hit = false;
    vec3 tc;
    float edge = 0.0;

    for (float t = incr; t <= maxRayDist && !hit && edge < 1.0; t += incr)
    {
        // view space ray march location
        vec3 p = pos + t * dir;

        // convert to screen texture coordinates
        tc = ssrViewToScreen(projection, p).xyz;

        // coarse sample
        float sampleDepth = fetchDepth(depthTex, tc.xy, nearFar, 0.0, size, ratio);

        // coarse intersection check
        bool wasUnder = under;
        under = p.z < sampleDepth;

        if (!wasUnder && under)
        {
            float best = sampleDepth - p.z;
            
            for (int i = 0; i < SSR_REFINE_STEPS; ++i)
            {
                float t2 = t - (1.0 - float(i + 1) / float(SSR_REFINE_STEPS + 1)) * incr;
                vec3 p2 = pos + t2 * dir;
                vec3 tc2 = ssrViewToScreen(projection, p2).xyz;

                float sampleDepth2 = fetchDepth(depthTex, tc2.xy, nearFar, 0.0, size, ratio);
                float diff = sampleDepth2 - p2.z;

                if( diff >= 0.0 && diff < best )
                {
                    best = diff;
                    tc = tc2;
                }
            }

            hit = best < (0.25 * incr);
        }

        // edge fade
        edge = max(abs(2.0 * tc.x - 1.0), abs(2.0 * tc.y - 1.0) );
    }

    // snap texcoords to nearest texel
    // tc = (floor(tc * uScreenColorSize.xy) + vec2(0.5,0.5)) * uScreenColorSize.zw;
    
    // output final values
    tc.z = 1.0 / tc.z;
    vec3 coords = hit ? tc.xyz : vec3(0.0);
    float mask = hit ? clamp(8.0 - 8.0 * edge, 0.0, 1.0) : 0.0;
    return vec4(coords, mask);
}

vec3 marmosetImportanceSampling(const in vec3 tangentX, const in vec3 tangentY, const in vec3 tangentZ,
                                const in vec3 eyeVector, const in float roughness, const in float jitter) {
    // full glossy
    // return normalize(reflect(-eyeVector, normal));

    float specExp = -10.0 / log2((1.0 - roughness) * 0.968 + 0.03);
    specExp *= specExp;

    // half vector is taken at random from probability distribution
    vec2 rnd = texture(uTextureBlueNoise, gl_FragCoord.xy / 64.0 + uHalton.xy).xy;

    vec4 r = vec4(rnd.x, rnd.y, cos(2.0 * 3.14159 * rnd.y), sin(2.0 * 3.14159 * rnd.y));
    r.zw = r.zw * 2.0 - vec2(1.0, 1.0);
    r.zw = normalize(r.zw);
    float cos_theta = pow(r.x, 1.0 / (specExp + 1.0));
    float sin_theta = sqrt(1.0 - cos_theta * cos_theta);
    float cos_phi = r.z;
    float sin_phi = r.w;
    vec3 h = vec3(cos_phi * sin_theta, sin_phi * sin_theta, cos_theta);
    h = h.x * tangentX + h.y * tangentY + h.z * tangentZ;

    // get reflection vector from half vector
    return reflect(-eyeVector, h);
}


float computeScreenRayOriginDir(in vec3 viewVertex, in vec3 rayDirView, in vec2 nearFar, in mat4 projection, out vec3 rayDirUV, out vec3 rayOriginUV) {
    // make ray length and clip it against the near plane (don't want to trace beyond visible)
    // float rayLen = (viewVertex.z + rayDirView.z * nearFar.y) > -nearFar.x ? (-nearFar.x - viewVertex.z) / rayDirView.z : nearFar.y;
    float rayLen = max(nearFar.y + viewVertex.z, -viewVertex.z - nearFar.x);

    rayOriginUV = ssrViewToScreen(projection, viewVertex.xyz);
    vec3 rayEndUV = ssrViewToScreen(projection, viewVertex.xyz + rayDirView * rayLen);

    rayOriginUV.z = 1.0 / rayOriginUV.z;
    rayEndUV.z = 1.0 / rayEndUV.z;

    rayDirUV = rayEndUV - rayOriginUV;

    // clip the line to the viewport edges
    float scaleMaxX = min(1.0, 0.99 * (1.0 - rayOriginUV.x) / max(1e-5, rayDirUV.x));
    float scaleMaxY = min(1.0, 0.99 * (1.0 - rayOriginUV.y) / max(1e-5, rayDirUV.y));
    float scaleMinX = min(1.0, 0.99 * rayOriginUV.x / max(1e-5, -rayDirUV.x));
    float scaleMinY = min(1.0, 0.99 * rayOriginUV.y / max(1e-5, -rayDirUV.y));
    rayDirUV *= min(scaleMaxX, scaleMaxY) * min(scaleMinX, scaleMinY);

    // make line advance faster if direction is closer to pixel edges (this avoids sampling the same pixel twice)
    // rayDirUV *= 1.0 / max(abs(rayDirUV.x), abs(rayDirUV.y)); // godot
    // rayDirUV *= 1.0 / sqrt(0.01 + clamp(dot(rayDirUV.xy, rayDirUV.xy), 0.0, 1.0)); // marmoset

    // depth tolerance
    return 0.5 * rayLen / float(SSR_STEPS);
}

float getStepOffset(const in float frameMod) {
    return (interleavedGradientNoise(gl_FragCoord.xy, frameMod) - 0.5) * SSR_RAND_RAY;
}

vec4 rayTraceUnrealSimple(
    const in vec3 viewVertex, const in vec3 rayDirView, const in mat4 projection,
    const in sampler2D depthTex, const in vec2 ratio, const in vec2 nearFar, const in float boxRadius, const in vec4 halton,
    const in float roughness, const in vec2 size, const in float frameMod) {

    float invNumSteps = 1.0 / float(SSR_STEPS);

    vec3 rayOriginUV, rayDirUV;
    float depthTolerance = computeScreenRayOriginDir(viewVertex.xyz, rayDirView, nearFar, projection, rayDirUV, rayOriginUV);
    vec4 result = vec4(rayOriginUV, 1.0);

    float sampleTime = getStepOffset(frameMod) * invNumSteps + invNumSteps;
    float lastDiff = 0.0;

    for (int i = 0; i < SSR_STEPS; i++) {
        result.xyz = rayOriginUV + rayDirUV * sampleTime;
        
        float level = roughness * (float(i) * 4.0 * invNumSteps) + HZB_LEVEL_OFFSET;
        float depth = fetchDepth(depthTex, result.xy, nearFar, level, size, ratio);

        float depthDiff = -1.0 / result.z - depth;
        bool hit = abs(depthDiff + depthTolerance) < depthTolerance;

        // Find more accurate hit using line segment intersection
        float timeLerp = clamp(lastDiff / (lastDiff - depthDiff), 0.0, 1.0);
        float hitTime = hit ? (sampleTime + timeLerp * invNumSteps - invNumSteps) : 1.0;
        result.w = min(result.w, hitTime);

        lastDiff = depthDiff;

        sampleTime += invNumSteps;
    }

    result.xyz = rayOriginUV + rayDirUV * result.w;
    result.w = 1.0 - result.w;
    return result;
}