#include "utils/octahedronLod.glsl"

uniform mat3 uEnvironmentTransform;
uniform vec3 uDiffuseSPH[9];

#define DFG_LUT_MULTISCATTER

#ifdef DFG_LUT_MULTISCATTER
uniform sampler2D uTextureDFG;
#endif

vec3 computeDiffuseSPH(const in vec3 normal) {
    float x = normal.x;
    float y = normal.y;
    float z = -normal.z;
    vec3 result = uDiffuseSPH[0] + uDiffuseSPH[1] * y + uDiffuseSPH[2] * z + uDiffuseSPH[3] * x +
                  uDiffuseSPH[4] * y * x + uDiffuseSPH[5] * y * z + uDiffuseSPH[6] * (3.0 * z * z - 1.0) +
                  uDiffuseSPH[7] * (z * x) + uDiffuseSPH[8] * (x * x - y * y);
    return max(result, vec3(0.0));
}

vec2 computeDFG(const in float roughness, const in float NoV) {
    // https://github.com/EpicGames/UnrealEngine/blob/release/Engine/Shaders/Private/BRDF.ush#L202
#ifdef DFG_LUT_MULTISCATTER
    return texture(uTextureDFG, vec2(NoV, roughness)).xy;
#else
    // https://www.unrealengine.com/blog/physically-based-shading-on-mobile
    const vec4 c0 = vec4(-1.0, -0.0275, -0.572, 0.022);
    const vec4 c1 = vec4(1.0, 0.0425, 1.04, -0.04);
    vec4 r = roughness * c0 + c1;
    float a004 = min(r.x * r.x, exp2(-9.28 * NoV)) * r.x + r.y;
    return vec2(-1.04, 1.04) * a004 + r.zw;
#endif
}

// https://seblagarde.files.wordpress.com/2015/07/course_notes_moving_frostbite_to_pbr_v32.pdf
vec3 getSpecularDominantDir(const in vec3 N, const in vec3 R, const in float roughness) {
    float a2 = roughness * roughness;
    return normalize(mix(N, R, (1.0 - a2) * (sqrt(1.0 - a2) + a2)));
}

vec3 computeIBLSpecularUE4(const in float NoV, const in vec3 normal, const in vec3 eyeVector, const in float roughness,
                           const in vec3 frontNormal) {
    vec3 R = normalize((2.0 * NoV) * normal - eyeVector);
    R = getSpecularDominantDir(normal, R, roughness);

    vec3 prefilteredColor = prefilterEnvMap(roughness, uEnvironmentTransform * R);

#ifdef HAS_NORMALMAP
    // http://marmosetco.tumblr.com/post/81245981087
    // marmoset uses 1.3, we force it to 1.0
    float factor = clamp(1.0 + dot(R, frontNormal), 0.0, 1.0);
    prefilteredColor *= factor * factor;
#endif

    return prefilteredColor;
}
