#include "shading/lightCommon.glsl"

// https://github.com/EpicGames/UnrealEngine/blob/release/Engine/Shaders/Private/BRDF.ush

vec4 precomputeGGX(const in float NoV, const in float roughness) {
    float a = roughness * roughness;
    return vec4(a, a * a, NoV, NoV * (1.0 - a));
}

// Walter et al. 2007, "Microfacet models for refraction through rough surfaces"
float D_GGX(const vec4 precomputeGGX, const float NoH) {
    float a2 = precomputeGGX.y;
    float d = (NoH * a2 - NoH) * NoH + 1.0;
    return a2 / (M_PI * d * d);
}

// Schlick 1994, "An Inexpensive BRDF Model for Physically-Based Rendering"
vec3 F_Schlick(const vec3 f0, const float f90, const in float VoH) {
    float VoH5 = pow(1.0 - VoH, 5.0);
    return f90 * VoH5 + (1.0 - VoH5) * f0;
}

// greyscale f0 version (clearcoat)
float F_Schlick(const float f0, const float f90, const in float VoH) { return f0 + (f90 - f0) * pow(1.0 - VoH, 5.0); }

// Heitz 2014, "Understanding the Masking-Shadowing Function in Microfacet-Based BRDFs"
float V_SmithCorrelated(const vec4 precomputeGGX, const float NoL) {
    float a = precomputeGGX.x;
    float smithV = NoL * (precomputeGGX.w + a);
    float smithL = precomputeGGX.z * (NoL * (1.0 - a) + a);
    return 0.5 / (smithV + smithL);
    // same as
    // Hammon 2017, "PBR Diffuse Lighting for GGX+Smith Microsurfaces"
    // return 0.5 / mix(2.0 * NoL * NoV, NoL + NoV, precomputeGGX.y)
}

vec3 specularLobe(const vec4 precomputeGGX, const vec3 normal, const vec3 eyeVector, const vec3 eyeLightDir,
                  const vec3 specular, const float NoL, const float f90) {
    vec3 H = normalize(eyeVector + eyeLightDir);
    float NoH = saturate(dot(normal, H));
    float VoH = saturate(dot(eyeLightDir, H));

    // distribution / visibility / fresnel
    float D = D_GGX(precomputeGGX, NoH);
    float V = V_SmithCorrelated(precomputeGGX, NoL);
    vec3 F = F_Schlick(specular, f90, VoH);

    return (D * V * M_PI) * F; // we multipy by PI because our diffuse is not divided by PI
}

// pure compute Light PBR
void computeLightLambertGGX(const in vec3 normal, const in vec3 eyeVector, const in float NoL,
                            const in vec4 precomputeGGX, const in vec3 diffuse, const in vec3 specular,
                            const in float attenuation, const in vec3 lightColor, const in vec3 eyeLightDir,
                            const in float f90, out vec3 diffuseOut, out vec3 specularOut, out bool lighted) {

    lighted = NoL > 0.0;
    if (lighted == false) {
        specularOut = diffuseOut = vec3(0.0);
        return;
    }

    vec3 colorAttenuate = attenuation * NoL * lightColor;
    specularOut = colorAttenuate * specularLobe(precomputeGGX, normal, eyeVector, eyeLightDir, specular, NoL, f90);
    diffuseOut = colorAttenuate * diffuse;
}
