#include "shading/environment.glsl"
#include "utils/eyeVector.glsl"
#include "utils/shadowing.glsl"

#ifdef HAS_LIGHT
#    include "shading/lighting.glsl"
#endif

#if defined(HAS_SHADOW) || defined(HAS_LIGHT_SPOT) || defined(HAS_LIGHT_POINT)
#    define NEED_EYE_POS
#endif

#if defined(NEED_EYE_POS) && !defined(FORWARD_VIEW_VERTEX)
#    include "utils/project.glsl"
#    include "utils/readDepth.glsl"

uniform sampler2D uTextureDepth;
#endif

#ifdef HAS_SSAO
uniform sampler2D uTextureSsaoFinal;
#endif

#ifdef HAS_SSR
uniform sampler2D uTextureSsrFinal;
uniform vec2 uTextureSsrFinalSize;
uniform float uSsrFactor;
#endif

uniform float uEnvironmentExposure;

// #define NORMAL_FILTERING
#define MIN_ROUGHNESS 0.045

// perceptualRoughness = matRoughness
// roughness = perceptualRoughness * perceptualRoughness

#ifdef NORMAL_FILTERING
float normalFiltering(float roughness, const vec3 normal) {
    // Kaplanyan 2016, "Stable specular highlights"
    // Tokuyoshi 2017, "Error Reduction and Simplification for Shading Anti-Aliasing"
    // Tokuyoshi and Kaplanyan 2019, "Improved Geometric Specular Antialiasing"

    // https://github.com/google/filament/blob/29a33ef21598e1013a2c355f2908ab4bd68bc11e/shaders/src/shading_lit.fs#L30
    float SIGMA2 = 0.15;
    float KAPPA = 0.2;

    vec3 dndu = dFdx(normal);
    vec3 dndv = dFdy(normal);
    float variance = SIGMA2 * (dot(dndu, dndu) + dot(dndv, dndv));
    float kernelRoughness2 = min(2.0 * variance, KAPPA);
    float filteredRoughness2 = saturate(roughness * roughness + kernelRoughness2);

    return sqrt(filteredRoughness2);
}
#endif

// ----------------------------------------------------------------
// ----------------------------------------------------------------

#ifdef HAS_SSAO

float specularOcclusion(const in float NoV, const in float ao, const in float roughness) {
    // https://github.com/google/filament/blob/b1070c6508ec50ab8c4a40a7f24b40da49c628cc/shaders/src/ambient_occlusion.fs#L77
    return saturate(pow(NoV + ao, exp2(-16.0 * roughness * roughness - 1.0)) - 1.0 + ao);
}

vec3 gtaoMultiBounce(float ao, const vec3 color) {
    // https://github.com/google/filament/blob/b1070c6508ec50ab8c4a40a7f24b40da49c628cc/shaders/src/ambient_occlusion.fs#L149
    // Jimenez et al. 2016, "Practical Realtime Strategies for Accurate Indirect Occlusion"
    vec3 a = 2.0404 * color - 0.3324;
    vec3 b = -4.7951 * color + 0.6417;
    vec3 c = 2.7552 * color + 0.6903;

    return max(vec3(ao), ((ao * a + b) * ao + c) * ao);
}
#endif

// ----------------------------------------------------------------
// ----------------------------------------------------------------

void shading_pbr(const in vec3 normal, const in vec3 baseColor, const in float roughIn, const in float metalness) {
    float f0 = 0.04;

    vec3 eyeVector = getEyeVector();

#ifdef NORMAL_FILTERING
    float roughness = normalFiltering(roughIn, normal);
#else
    float roughness = roughIn;
#endif

    // ----------------------------------------------------------------
    // -- material
    // ----------------------------------------------------------------

    vec3 matDiffuse = baseColor * ((1.0 - metalness) * (1.0 - f0));
    vec3 matSpecular = baseColor * metalness + f0 * (1.0 - metalness);
    float matRoughness = max(MIN_ROUGHNESS, roughness);
    float f90 = clamp(50.0 * matSpecular.g, 0.0, 1.0);
    float NoV = saturate(dot(normal, eyeVector));

    // ----------------------------------------------------------------
    // -- environment
    // ----------------------------------------------------------------

    vec2 dfg = computeDFG(matRoughness, NoV);
    vec3 dfgSpec = mix(dfg.xxx * f90, dfg.yyy, matSpecular);

    g_diffuse = matDiffuse * computeDiffuseSPH(uEnvironmentTransform * normal);
    g_specular = computeIBLSpecularUE4(NoV, normal, eyeVector, matRoughness, normal);

    g_diffuse *= uEnvironmentExposure * (1.0 - dfgSpec);
    g_specular *= uEnvironmentExposure * dfgSpec;

    vec3 energyCompensation = 1.0 + matSpecular * (1.0 / dfg.y - 1.0);

    // ----------------------------------------------------------------
    // -- ssao
    // ----------------------------------------------------------------

#ifdef HAS_SSAO
    {
        float matAO = texture(uTextureSsaoFinal, g_uv).r;
        // g_diffuse *= matAO;
        // g_specular *= specularOcclusion(NoV, matAO, matRoughness);

        g_diffuse *= gtaoMultiBounce(matAO, matDiffuse);
        g_specular *= gtaoMultiBounce(specularOcclusion(NoV, matAO, matRoughness), matSpecular);
    }
#endif

    // ----------------------------------------------------------------
    // ---- ssr
    // ----------------------------------------------------------------

#ifdef HAS_SSR
    {
        vec4 ssr = texture(uTextureSsrFinal, g_uv);
        g_specular = mix(g_specular, matSpecular * ssr.rgb / energyCompensation, uSsrFactor * ssr.w);
    }
#endif

    // ----------------------------------------------------------------
    // ---- light
    // ----------------------------------------------------------------

    // light
#ifdef HAS_LIGHT
    float attenuation, dotNL;
    vec3 eyeLightDir;
    bool lighted;
    vec3 lightSpecular;
    vec3 lightDiffuse;
    vec4 prepGGX = precomputeGGX(NoV, matRoughness);
#endif

#ifdef NEED_EYE_POS
    float shadow;
#    ifdef FORWARD_VIEW_VERTEX
    vec3 eyePos = FORWARD_VIEW_VERTEX;
#    else
    vec3 eyePos = eyeVector * depthLinearToZEye(readDepthLinear(texture(uTextureDepth, g_uv).x));
#    endif
#endif

#ifdef HAS_SHADOW
    vec3 shadowUV;
#endif

#REPEAT 3
#ifdef HAS_LIGHT_$ID$

    // light type
#    if LIGHT_TYPE_$ID$ == LIGHT_DIRECTIONAL
    precomputeDirectional(normal, uLightViewDirection$ID$, attenuation, eyeLightDir, dotNL);

#    elif LIGHT_TYPE_$ID$ == LIGHT_SPOT
    precomputeSpot(normal, eyePos, uLightViewDirection$ID$, uLightViewPosition$ID$, uLightSpotAngle$ID$,
                   uLightSpotSoftness$ID$, attenuation, eyeLightDir, dotNL);

#    elif LIGHT_TYPE_$ID$ == LIGHT_POINT

    precomputePoint(normal, eyePos, uLightViewPosition$ID$, attenuation, eyeLightDir, dotNL);
#    endif

    // shading
    computeLightLambertGGX(normal, eyeVector, dotNL, prepGGX, matDiffuse, matSpecular, attenuation, uLightColor$ID$.rgb,
                           eyeLightDir, f90, lightDiffuse, lightSpecular, lighted);

    // shadowing
#    ifdef HAS_SHADOW_$ID$
#        if LIGHT_TYPE_$ID$ == LIGHT_DIRECTIONAL
    shadowUV = getDirectionalUV(eyePos, normal, dotNL, uShadowNormalBias$ID$, uShadowProject$ID$);
#        elif LIGHT_TYPE_$ID$ == LIGHT_SPOT
    shadowUV = getSpotUV(eyePos, normal, dotNL, uShadowNormalBias$ID$, uShadowProject$ID$, uShadowNearFar$ID$);
#        endif

    shadow = shadowGet(uTextureShadow$ID$, shadowUV.xyz);
    lightDiffuse *= shadow;
    lightSpecular *= shadow;
#    endif

    g_diffuse += lightDiffuse;
    g_specular += lightSpecular;

    // // debug shadow
    // energyCompensation *= 0.0;
    // g_diffuse = shadowUV.zzz;
    // if (g_uv.x > 0.5) g_diffuse = texture(uTextureShadow0, shadowUV.xy).xxx;
    // g_diffuse = sRGBToLinear(g_diffuse);

#endif
#REPEAT END

    // ----------------------------------------------------------------
    // ----------------------------------------------------------------

    // energy compensation
    g_specular *= energyCompensation;
}
