#define SHADOW_VSM

uniform vec2 uShadowSize;
uniform float uFrameJitter;
#include "utils/rand.glsl"

#ifdef SHADOW_VSM
#    define shadowGet shadowVsm
#else
#    define shadowGet shadowPcf
#endif

// ---------------------------------------------------------

vec3 getDirectionalUV(const vec3 viewPos, const vec3 viewNormal, const in float dotNL, const in float normalBias,
                      const in mat4 uvFromViewMatrix) {
    vec3 viewPosBias = viewPos + viewNormal * (sqrt(1.0 - dotNL * dotNL) * normalBias);
    return 0.5 + 0.5 * (uvFromViewMatrix * vec4(viewPosBias, 1.0)).xyz;
}

vec3 getSpotUV(const vec3 viewPos, const vec3 viewNormal, const in float dotNL, const in float normalBias,
               const in mat4 uvFromViewMatrix, const in vec2 nearFar) {
    vec3 viewPosBias = viewPos + viewNormal * (sqrt(1.0 - dotNL * dotNL) * normalBias);
    vec3 pos = (uvFromViewMatrix * vec4(viewPosBias, 1.0)).xyw;
    pos.xy = 0.5 + 0.5 * pos.xy / pos.z;
    pos.z = (pos.z - nearFar.x) / (nearFar.y - nearFar.x);
    return pos;
}

// ---------------------------------------------------------

#ifdef SHADOW_VSM

float linstep(const float a, const float b, const float v) { return clamp((v - a) / (b - a), 0.0, 1.0); }

float chebyshevUpperBound(const vec2 moments, const float mean) {
    // Donnelly and Lauritzen 2006, "Variance Shadow Maps"

    const float minVariance = 0.00001;
    const float lightBleedReduction = 0.2;

    float variance = moments.y - (moments.x * moments.x);
    variance = max(variance, minVariance);

    float d = mean - moments.x;
    float pMax = variance / (variance + d * d);
    pMax = linstep(lightBleedReduction, 1.0, pMax);

    return mean <= moments.x ? 1.0 : pMax;
}

float shadowVsm(const in sampler2D tex, const vec3 posUV) {
    // if (posUV.x <= 0.0 || posUV.x >= 1.0) return 1.0;
    // if (posUV.y <= 0.0 || posUV.y >= 1.0) return 1.0;
    // if (posUV.z <= 0.0 || posUV.z >= 1.0) return 1.0;

    float rnd = 0.0;
    if (uFrameJitter > 0.0) {
        rnd = interleaveGradientNoise(gl_FragCoord.xy, uFrameJitter) * 2.0 - 1.0;
    }

    vec2 moments = texture(tex, posUV.xy + rnd / uShadowSize).xy;
    float depth = min(posUV.z, 1.0);

    return chebyshevUpperBound(moments, depth);
}

#else

// ---------------------------------------------------------

float shadowPcf(const in sampler2D tex, const vec3 posUV) {
    vec2 texelSize = 1.0 / uShadowSize;

    vec2 offset = vec2(0.5);
    vec2 uv = clamp(posUV.xy, -1.0, 2.0) * uShadowSize + offset;
    vec2 base = (floor(uv) - offset) * texelSize;
    vec2 st = fract(uv);

    vec2 uw = vec2(3.0 - 2.0 * st.x, 1.0 + 2.0 * st.x);
    vec2 vw = vec2(3.0 - 2.0 * st.y, 1.0 + 2.0 * st.y);

    vec2 u = texelSize.x * vec2((2.0 - st.x) / uw.x - 1.0, st.x / uw.y + 1.0);
    vec2 v = texelSize.y * vec2((2.0 - st.y) / vw.x - 1.0, st.y / vw.y + 1.0);

    float sum = 0.0;

    sum += uw.x * vw.x * texture(tex, base + vec2(u.x, v.x)).r;
    sum += uw.y * vw.x * texture(tex, base + vec2(u.y, v.x)).r;

    sum += uw.x * vw.y * texture(tex, base + vec2(u.x, v.y)).r;
    sum += uw.y * vw.y * texture(tex, base + vec2(u.y, v.y)).r;

    return sum * (1.0 / 16.0);
}

#endif
